import sempy.fabric as fabric
import pandas as pd
from typing import Optional
import sempy_labs._icons as icons
from uuid import UUID
from .._helper_functions import (
    resolve_workspace_name_and_id,
)
from sempy._utils._log import log


@log
def get_direct_lake_guardrails() -> pd.DataFrame:
    """
    Shows the guardrails for when Direct Lake semantic models will fallback to Direct Query
    based on Microsoft's `online documentation <https://learn.microsoft.com/power-bi/enterprise/directlake-overview>`_.

    Returns
    -------
    pandas.DataFrame
        A table showing the Direct Lake guardrails by SKU.
    """

    url = "https://learn.microsoft.com/power-bi/enterprise/directlake-overview"

    tables = pd.read_html(url)
    for df in tables:
        first_column_name = df.columns[0]
        if first_column_name.startswith("Fabric"):
            df[first_column_name] = df[first_column_name].str.split("/")
            df = df.explode(first_column_name, ignore_index=True)
            break

    return df


def get_sku_size(workspace: Optional[str | UUID] = None) -> str:
    """
    Shows the SKU size for a workspace.

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    str
        The SKU size for a workspace.
    """
    from sempy_labs._capacities import list_capacities

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    dfW = fabric.list_workspaces(filter=f"id eq '{workspace_id}'")

    if len(dfW) == 0:
        raise ValueError(
            f"{icons.red_dot} The '{workspace_name}' is not a valid workspace."
        )

    capacity_id = dfW["Capacity Id"].iloc[0]
    dfC = list_capacities()
    dfC_filt = dfC[dfC["Id"] == capacity_id]

    if len(dfC_filt) == 0:
        raise ValueError(
            f"{icons.red_dot} The '{capacity_id}' Id is not a valid capacity Id."
        )

    return dfC_filt["Sku"].iloc[0]


def get_directlake_guardrails_for_sku(sku_size: str) -> pd.DataFrame:
    """
    Shows the guardrails for Direct Lake based on the SKU used by your workspace's capacity.
    * Use the result of the 'get_sku_size' function as an input for this function's sku_size parameter.*

    Parameters
    ----------
    sku_size : str
        Sku size of a workspace/capacity.

    Returns
    -------
    pandas.DataFrame
        A table showing the Direct Lake guardrails for the given SKU.
    """

    df = get_direct_lake_guardrails()
    col_name = df.columns[0]
    filtered_df = df[df[col_name] == sku_size]

    return filtered_df



from uuid import UUID
from typing import Optional
from .._helper_functions import (
    resolve_workspace_name_and_id,
    resolve_item_name_and_id,
    _base_api,
)
import sempy_labs._icons as icons
from sempy._utils._log import log


@log
def refresh_catalog_metadata(
    mirrored_azure_databricks_catalog: str | UUID,
    workspace: Optional[str | UUID] = None,
):
    """
    Refresh Databricks catalog metadata in mirroredAzureDatabricksCatalogs Item.

    This is a wrapper function for the following API: `Refresh Metadata - Items RefreshCatalogMetadata <https://learn.microsoft.com/rest/api/fabric/mirroredazuredatabrickscatalog/refresh-metadata/items-refresh-catalog-metadata>`_.

    Parameters
    ----------
    mirrored_azure_databricks_catalog : str | uuid.UUID
        The name or ID of the mirrored Azure Databricks catalog.
    workspace : str | uuie.UUID, default=None
        The workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (catalog_name, catalog_id) = resolve_item_name_and_id(
        mirrored_azure_databricks_catalog
    )

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/mirroredAzureDatabricksCatalogs/{catalog_id}/refreshCatalogMetadata",
        method="post",
        lro_return_status_code=True,
    )

    print(
        f"{icons.green_dot} The '{catalog_name}' Databricks Catalog metadata within the '{workspace_name}' workspace has been refreshed."
    )



from .._helper_functions import (
    _base_api,
    _build_url,
    _encode_user,
    _update_dataframe_datatypes,
    _create_dataframe,
)

from uuid import UUID
from typing import Optional
from ._basic_functions import (
    _resolve_workspace_name_and_id,
)
import sempy_labs._icons as icons
import pandas as pd
from sempy._utils._log import log


@log
def add_user_to_workspace(
    user: str | UUID,
    role: str = "Member",
    principal_type: str = "User",
    workspace: Optional[str | UUID] = None,
):
    """
    Grants user permissions to the specified workspace.

    This is a wrapper function for the following API: `Admin - Groups AddUserAsAdmin <https://learn.microsoft.com/rest/api/power-bi/admin/groups-add-user-as-admin>`_.

    Parameters
    ----------
    user : str | uuid.UUID
        The user identifier or email address. For service principals and groups you must use the user identifier.
    role : str, default="Member"
        The role of the user in the workspace. Options are: 'Admin', 'Contributor', 'Member', 'None', 'Viewer'.
    principal_type : str, default="User"
        The principal type of the user. Options are: 'App', 'Group', 'None', 'User'.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = _resolve_workspace_name_and_id(workspace)

    # Validation
    role = role.capitalize()
    roles = ["Admin", "Contributor", "Member", "None", "Viewer"]
    if role not in roles:
        raise ValueError(f"{icons.red_dot} Invalid role. Please choose from {roles}")
    principal_types = ["App", "Group", "None", "User"]
    if principal_type not in principal_types:
        raise ValueError(
            f"{icons.red_dot} Invalid principal type. Please choose from {principal_types}"
        )

    user = _encode_user(user)

    payload = {
        "identifier": user,  # identifier or emailAddress?
        "principalType": principal_type,
        "groupUserAccessRight": role,
    }

    _base_api(
        request=f"/v1.0/myorg/admin/groups/{workspace_id}/users",
        method="post",
        payload=payload,
    )

    print(
        f"{icons.green_dot} The '{user}' user has been added with '{role.lower()}' permissions to the '{workspace_name}' workspace."
    )


@log
def delete_user_from_workspace(
    user: str | UUID,
    workspace: Optional[str | UUID] = None,
    is_group: Optional[bool] = None,
    profile_id: Optional[str] = None,
):
    """
    Removes user permissions from the specified workspace.

    This is a wrapper function for the following API: `Admin - Groups DeleteUserAsAdmin <https://learn.microsoft.com/rest/api/power-bi/admin/groups-delete-user-as-admin>`_.

    Parameters
    ----------
    user : str | uuid.UUID
        The user identifier or email address. For service principals and groups you must use the user identifier.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    is_group : bool, default=None
        Whether a given user is a group or not. This parameter is required when user to delete is group.
    profile_id : str, default=None
        The service principal profile ID to delete.
    """

    (workspace_name, workspace_id) = _resolve_workspace_name_and_id(workspace)

    user = _encode_user(user)
    url = f"/v1.0/myorg/admin/groups/{workspace_id}/users/{user}"

    params = {}
    if profile_id is not None:
        params["profileId"] = profile_id
    if is_group is not None:
        params["isGroup"] = is_group

    url = _build_url(url, params)

    _base_api(
        request=url,
        method="delete",
    )

    print(
        f"{icons.green_dot} The '{user}' user has been removed from the '{workspace_name}' workspace."
    )


@log
def restore_deleted_workspace(workspace_id: UUID, name: str, email_address: str):
    """
    Restores a deleted workspace.

    This is a wrapper function for the following API: `Admin - Groups RestoreDeletedGroupAsAdmin <https://learn.microsoft.com/rest/api/power-bi/admin/groups-restore-deleted-group-as-admin>`_.

    Parameters
    ----------
    workspace_id : uuid.UUID
        The ID of the workspace to restore.
    name : str
        The name of the group to be restored
    email_address : str
        The email address of the owner of the group to be restored
    """

    payload = {
        "name": name,
        "emailAddress": email_address,
    }

    _base_api(
        request=f"/v1.0/myorg/admin/groups/{workspace_id}/restore",
        method="post",
        payload=payload,
    )

    print(
        f"{icons.green_dot} The '{workspace_id}' workspace has been restored as '{name}'."
    )


@log
def list_orphaned_workspaces(top: int = 100) -> pd.DataFrame:
    """
    Shows a list of orphaned workspaces (those with no users or no admins).

    This is a wrapper function for the following API:
    `Admin - Groups ListGroupsAsAdmin <https://learn.microsoft.com/rest/api/power-bi/admin/groups-get-groups-as-admin>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    top : int, default=100
        The maximum number of results to return.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of orphaned workspaces.
    """

    # column structure with proper data types
    columns = {
        "Workspace Name": "string",
        "Workspace Id": "string",
        "Type": "string",
        "State": "string",
        "Is Read Only": "bool",
        "Is On Dedicated Capacity": "bool",
        "Capacity Migration Status": "string",
        "Has Workspace Level Settings": "bool",
        "Users": "list",
    }

    df = _create_dataframe(columns=columns)

    url = (
        "/v1.0/myorg/admin/groups?"
        "$expand=users&"
        "$filter=(not users/any()) or "
        "(not users/any(u: u/groupUserAccessRight eq Microsoft.PowerBI.ServiceContracts.Api.GroupUserAccessRight'Admin'))&"
        f"$top={top}"
    )

    response = _base_api(request=url, client="fabric_sp")
    values = response.json().get("value", [])
    df_raw = pd.json_normalize(values)

    # friendly names and reorder
    if not df_raw.empty:
        df_raw = df_raw.rename(
            columns={
                "name": "Workspace Name",
                "id": "Workspace Id",
                "type": "Type",
                "state": "State",
                "isReadOnly": "Is Read Only",
                "isOnDedicatedCapacity": "Is On Dedicated Capacity",
                "capacityMigrationStatus": "Capacity Migration Status",
                "hasWorkspaceLevelSettings ": "Has Workspace Level Settings",  # Note the space in original
                "users": "Users",
            }
        )

        df = df_raw[list(columns.keys())].copy()

        # Convert empty lists to a more readable format for Users column
        if "Users" in df.columns:
            df["Users"] = df["Users"].apply(
                lambda x: x if (x is not None and len(x) > 0) else []
            )
    else:
        df = _create_dataframe(columns=columns)

    # proper data types
    _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df



import sempy.fabric as fabric
import pandas as pd
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    format_dax_object_name,
    resolve_dataset_name_and_id,
    _base_api,
    generate_guid,
)
from sempy_labs._model_dependencies import get_model_calc_dependencies
from typing import Optional, List, Tuple
from sempy._utils._log import log
from uuid import UUID
from sempy_labs.directlake._warm_cache import _put_columns_into_memory
import sempy_labs._icons as icons
import time


@log
def evaluate_dax_impersonation(
    dataset: str | UUID,
    dax_query: str,
    user_name: Optional[str] = None,
    workspace: Optional[str | UUID] = None,
):
    """
    Runs a DAX query against a semantic model using the `REST API <https://learn.microsoft.com/en-us/rest/api/power-bi/datasets/execute-queries-in-group>`_.

    Compared to evaluate_dax this allows passing the user name for impersonation.
    Note that the REST API has significant limitations compared to the XMLA endpoint.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    dax_query : str
        The DAX query.
    user_name : str
        The user name (i.e. hello@goodbye.com).
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe holding the result of the DAX query.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    payload = {
        "queries": [{"query": dax_query}],
        "impersonatedUserName": user_name,
    }

    response = _base_api(
        request=f"/v1.0/myorg/groups/{workspace_id}/datasets/{dataset_id}/executeQueries",
        method="post",
        payload=payload,
    )
    data = response.json()["results"][0]["tables"]

    # Get all possible column names from all rows because null columns aren't returned
    all_columns = set()
    for item in data:
        for row in item["rows"]:
            all_columns.update(row.keys())

    # Create rows with all columns, filling missing values with None
    rows = []
    for item in data:
        for row in item["rows"]:
            # Create a new row with all columns, defaulting to None
            new_row = {col: row.get(col) for col in all_columns}
            rows.append(new_row)

    # Create DataFrame from the processed rows
    df = pd.DataFrame(rows)

    return df


@log
def get_dax_query_dependencies(
    dataset: str | UUID,
    dax_string: str | List[str],
    put_in_memory: bool = False,
    show_vertipaq_stats: bool = True,
    workspace: Optional[str | UUID] = None,
) -> pd.DataFrame:
    """
    Obtains the columns on which a DAX query depends, including model dependencies. Shows Vertipaq statistics (i.e. Total Size, Data Size, Dictionary Size, Hierarchy Size) for easy prioritizing.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    dax_string : str | List[str]
        The DAX query or list of DAX queries.
    put_in_memory : bool, default=False
        If True, ensures that the dependent columns are put into memory in order to give realistic Vertipaq stats (i.e. Total Size etc.).
    show_vertipaq_stats : bool, default=True
        If True, shows vertipaq stats (i.e. Total Size, Data Size, Dictionary Size, Hierarchy Size)
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the dependent columns of a given DAX query including model dependencies.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    fabric.refresh_tom_cache(workspace=workspace)

    if isinstance(dax_string, str):
        dax_string = [dax_string]

    final_df = pd.DataFrame(columns=["Object Type", "Table", "Object"])

    cd = get_model_calc_dependencies(dataset=dataset_id, workspace=workspace_id)

    for dax in dax_string:
        # Escape quotes in dax
        dax = dax.replace('"', '""')
        final_query = f"""
            EVALUATE
            VAR source_query = "{dax}"
            VAR all_dependencies = SELECTCOLUMNS(
                INFO.CALCDEPENDENCY("QUERY", source_query),
                    "Referenced Object Type",[REFERENCED_OBJECT_TYPE],
                    "Referenced Table", [REFERENCED_TABLE],
                    "Referenced Object", [REFERENCED_OBJECT]
                )             
            RETURN all_dependencies
            """
        dep = fabric.evaluate_dax(
            dataset=dataset_id, workspace=workspace_id, dax_string=final_query
        )

        # Clean up column names and values (remove outside square brackets, underscorees in object type)
        dep.columns = dep.columns.map(lambda x: x[1:-1])
        dep["Referenced Object Type"] = (
            dep["Referenced Object Type"].str.replace("_", " ").str.title()
        )

        # Dataframe df will contain the output of all dependencies of the objects used in the query
        df = dep.copy()

        for _, r in dep.iterrows():
            ot = r["Referenced Object Type"]
            object_name = r["Referenced Object"]
            table_name = r["Referenced Table"]
            cd_filt = cd[
                (cd["Object Type"] == ot)
                & (cd["Object Name"] == object_name)
                & (cd["Table Name"] == table_name)
            ]

            # Adds in the dependencies of each object used in the query (i.e. relationship etc.)
            if len(cd_filt) > 0:
                subset = cd_filt[
                    ["Referenced Object Type", "Referenced Table", "Referenced Object"]
                ]
                df = pd.concat([df, subset], ignore_index=True)

        df.columns = df.columns.map(lambda x: x.replace("Referenced ", ""))
        final_df = pd.concat([df, final_df], ignore_index=True)

    final_df = final_df[
        (final_df["Object Type"].isin(["Column", "Calc Column"]))
        & (~final_df["Object"].str.startswith("RowNumber-"))
    ]
    final_df = final_df.drop_duplicates().reset_index(drop=True)
    final_df = final_df.rename(columns={"Table": "Table Name", "Object": "Column Name"})
    final_df.drop(columns=["Object Type"], inplace=True)

    if not show_vertipaq_stats:
        return final_df

    # Get vertipaq stats, filter to just the objects in the df dataframe
    final_df["Full Object"] = format_dax_object_name(
        final_df["Table Name"], final_df["Column Name"]
    )
    dfC = fabric.list_columns(dataset=dataset_id, workspace=workspace_id, extended=True)
    dfC["Full Object"] = format_dax_object_name(dfC["Table Name"], dfC["Column Name"])

    dfC_filtered = dfC[dfC["Full Object"].isin(final_df["Full Object"].values)][
        [
            "Table Name",
            "Column Name",
            "Total Size",
            "Data Size",
            "Dictionary Size",
            "Hierarchy Size",
            "Is Resident",
            "Full Object",
        ]
    ].reset_index(drop=True)

    if put_in_memory:
        # Only put columns in memory if they are in a Direct Lake table (and are not already in memory)
        dfP = fabric.list_partitions(dataset=dataset, workspace=workspace)
        dl_tables = dfP[dfP["Mode"] == "DirectLake"]["Table Name"].unique().tolist()
        not_in_memory = dfC_filtered[
            (dfC_filtered["Table Name"].isin(dl_tables))
            & (dfC_filtered["Is Resident"] == False)
        ]

        if not not_in_memory.empty:
            _put_columns_into_memory(
                dataset=dataset,
                workspace=workspace,
                col_df=dfC_filtered,
                return_dataframe=False,
            )

            # Get column stats again
            dfC = fabric.list_columns(
                dataset=dataset_id, workspace=workspace_id, extended=True
            )
            dfC["Full Object"] = format_dax_object_name(
                dfC["Table Name"], dfC["Column Name"]
            )

            dfC_filtered = dfC[dfC["Full Object"].isin(final_df["Full Object"].values)][
                [
                    "Table Name",
                    "Column Name",
                    "Total Size",
                    "Data Size",
                    "Dictionary Size",
                    "Hierarchy Size",
                    "Is Resident",
                    "Full Object",
                ]
            ].reset_index(drop=True)

    dfC_filtered.drop(["Full Object"], axis=1, inplace=True)

    return dfC_filtered


@log
def get_dax_query_memory_size(
    dataset: str | UUID, dax_string: str, workspace: Optional[str | UUID] = None
) -> int:
    """
    Obtains the total size, in bytes, used by all columns that a DAX query depends on.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    dax_string : str
        The DAX query.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    int
        The total size, in bytes, used by all columns that the DAX query depends on.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    df = get_dax_query_dependencies(
        dataset=dataset_id,
        workspace=workspace_id,
        dax_string=dax_string,
        put_in_memory=True,
    )

    return df["Total Size"].sum()


@log
def _dax_perf_test(
    dataset: str,
    dax_queries: dict,
    clear_cache_before_run: bool = False,
    refresh_type: Optional[str] = None,
    rest_time: int = 2,
    workspace: Optional[str] = None,
) -> Tuple[pd.DataFrame, dict]:
    """
    Runs a performance test on a set of DAX queries.

    Parameters
    ----------
    dataset : str
        Name of the semantic model.
    dax_queries : dict
        The dax queries to run in a dictionary format. Here is an example:
        {
            "Sales Amount Test", """ """ EVALUATE SUMMARIZECOLUMNS("Sales Amount", [Sales Amount]) """ """,
            "Order Quantity with Product", """ """ EVALUATE SUMMARIZECOLUMNS('Product'[Color], "Order Qty", [Order Qty]) """ """,
        }
    clear_cache_before_run : bool, default=False
    refresh_type : str, default=None
    rest_time : int, default=2
        Rest time (in seconds) between the execution of each DAX query.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    Tuple[pandas.DataFrame, dict]
        A pandas dataframe showing the SQL profiler trace results of the DAX queries.
        A dictionary of the query results in pandas dataframes.
    """
    from sempy_labs._refresh_semantic_model import refresh_semantic_model
    from sempy_labs._clear_cache import clear_cache

    event_schema = {
        "QueryBegin": [
            "EventClass",
            "EventSubclass",
            "CurrentTime",
            "NTUserName",
            "TextData",
            "StartTime",
            "ApplicationName",
        ],
        "QueryEnd": [
            "EventClass",
            "EventSubclass",
            "CurrentTime",
            "NTUserName",
            "TextData",
            "StartTime",
            "EndTime",
            "Duration",
            "CpuTime",
            "Success",
            "ApplicationName",
        ],
        "VertiPaqSEQueryBegin": [
            "EventClass",
            "EventSubclass",
            "CurrentTime",
            "NTUserName",
            "TextData",
            "StartTime",
        ],
        "VertiPaqSEQueryEnd": [
            "EventClass",
            "EventSubclass",
            "CurrentTime",
            "NTUserName",
            "TextData",
            "StartTime",
            "EndTime",
            "Duration",
            "CpuTime",
            "Success",
        ],
        "VertiPaqSEQueryCacheMatch": [
            "EventClass",
            "EventSubclass",
            "CurrentTime",
            "NTUserName",
            "TextData",
        ],
    }

    # Add Execution Metrics
    event_schema["ExecutionMetrics"] = ["EventClass", "ApplicationName", "TextData"]
    # Add DAX Query Plan
    # event_schema["DAXQueryPlan"] = ["EventClass", "EventSubclass", "CurrentTime", "StartTime", "EndTime", "Duration", "CpuTime", "ApplicationName", "TextData"]

    query_results = {}

    # Establish trace connection
    with fabric.create_trace_connection(
        dataset=dataset, workspace=workspace
    ) as trace_connection:
        with trace_connection.create_trace(event_schema) as trace:
            trace.start()
            print(f"{icons.in_progress} Starting performance testing...")
            # Loop through DAX queries
            for name, dax in dax_queries.items():

                if clear_cache_before_run:
                    clear_cache(dataset=dataset, workspace=workspace)
                if refresh_type is not None:
                    refresh_semantic_model(
                        dataset=dataset, workspace=workspace, refresh_type=refresh_type
                    )

                # EVALUATE {1} is used to initate a warm cache
                fabric.evaluate_dax(
                    dataset=dataset, workspace=workspace, dax_string="""EVALUATE {1}"""
                )
                # Run DAX Query
                result = fabric.evaluate_dax(
                    dataset=dataset, workspace=workspace, dax_string=dax
                )

                # Add results to output
                query_results[name] = result

                time.sleep(rest_time)
                print(f"{icons.green_dot} The '{name}' query has completed.")

            df = trace.stop()
            # Allow time to collect trace results
            time.sleep(5)

            # Step 1: Filter out unnecessary operations
            query_names = list(dax_queries.keys())
            df = df[
                ~df["Application Name"].isin(["PowerBI", "PowerBIEIM"])
                & (~df["Text Data"].str.startswith("EVALUATE {1}"))
            ]
            query_begin = df["Event Class"] == "QueryBegin"
            temp_column_name = "QueryName_INT"
            df = df.copy()
            df[temp_column_name] = query_begin.cumsum()
            df[temp_column_name] = (
                df[temp_column_name]
                .where(query_begin, None)  # Assign None to non-query begin rows
                .ffill()  # Forward fill None values
                .astype("Int64")  # Use pandas nullable integer type for numeric indices
            )

            df.loc[df[temp_column_name].notna(), "Query Name"] = (
                df[temp_column_name]
                .dropna()
                .astype(int)
                .map(lambda x: query_names[x - 1])
            )
            df = df[df[temp_column_name] != None]
            df = df.drop(columns=[temp_column_name])

            query_to_guid = {
                name: generate_guid() for name in df["Query Name"].unique()
            }
            df["Query ID"] = df["Query Name"].map(query_to_guid)

    df = df.reset_index(drop=True)

    return df, query_results


def _dax_perf_test_bulk(
    mapping: dict,
    clear_cache_before_run: bool = False,
    refresh_type: Optional[str] = None,
    rest_time: int = 2,
):
    """
    mapping is something like this:

    mapping = {
        "Workspace1": {
            "Dataset1": {
                "Query1": "EVALUATE ...",
                "Query2": "EVALUATE ...",
            },
            "Dataset2": {
                "Query3": "EVALUATE ...",
                "Query4": "EVALUATE ...",
            }
        },
        "Workspace2": {
            "Dataset3": {
                "Query5": "EVALUATE ...",
                "Query6": "EVALUATE ...",
            },
            "Dataset4": {
                "Query7": "EVALUATE ...",
                "Query8": "EVALUATE ...",
            }
        }
    }
    """

    for workspace, datasets in mapping.items():
        for dataset, queries in datasets.items():
            _dax_perf_test(
                dataset=dataset,
                dax_queries=queries,
                clear_cache_before_run=clear_cache_before_run,
                refresh_type=refresh_type,
                rest_time=rest_time,
                workspace=workspace,
            )



import pandas as pd
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    _is_valid_uuid,
    _update_dataframe_datatypes,
    _base_api,
    _create_dataframe,
    resolve_workspace_name,
    resolve_workspace_id,
    _decode_b64,
    _conv_b64,
    get_jsonpath_value,
)
from typing import Optional, Tuple
import sempy_labs._icons as icons
from uuid import UUID
from jsonpath_ng.ext import parse
import json


def list_dataflows(workspace: Optional[str | UUID] = None):
    """
    Shows a list of all dataflows which exist within a workspace.

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the dataflows which exist within a workspace.
    """

    workspace_id = resolve_workspace_id(workspace)

    columns = {
        "Dataflow Id": "string",
        "Dataflow Name": "string",
        "Description": "string",
        "Configured By": "string",
        "Users": "string",
        "Generation": "string",
    }
    df = _create_dataframe(columns=columns)

    response = _base_api(
        request=f"/v1.0/myorg/groups/{workspace_id}/dataflows", client="fabric_sp"
    )

    rows = []
    for v in response.json().get("value", []):
        gen = v.get("generation")
        rows.append(
            {
                "Dataflow Id": v.get("objectId"),
                "Dataflow Name": v.get("name"),
                "Description": "",
                "Configured By": v.get("configuredBy"),
                "Users": ", ".join(v.get("users", [])),
                "Generation": "Gen2" if gen == 2 else "Gen1",
            }
        )

    responses = _base_api(
        request=f"/v1/workspaces/{workspace_id}/dataflows",
        client="fabric_sp",
        uses_pagination=True,
    )
    for r in responses:
        for v in r.get("value", []):
            gen = v.get("generation")
            rows.append(
                {
                    "Dataflow Id": v.get("id"),
                    "Dataflow Name": v.get("displayName"),
                    "Description": v.get("description"),
                    "Configured By": "",
                    "Users": "",
                    "Generation": "Gen2 CI/CD",
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


def assign_workspace_to_dataflow_storage(
    dataflow_storage_account: str, workspace: Optional[str | UUID] = None
):
    """
    Assigns a dataflow storage account to a workspace.

    This is a wrapper function for the following API: `Dataflow Storage Accounts - Groups AssignToDataflowStorage <https://learn.microsoft.com/rest/api/power-bi/dataflow-storage-accounts/groups-assign-to-dataflow-storage>`_.

    Parameters
    ----------
    dataflow_storage_account : str
        The name of the dataflow storage account.
    workspace : str | uuid.UUID, default=None
        The name or ID of the workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    df = list_dataflow_storage_accounts()
    df_filt = df[df["Dataflow Storage Account Name"] == dataflow_storage_account]

    if len(df_filt) == 0:
        raise ValueError(
            f"{icons.red_dot} The '{dataflow_storage_account}' does not exist."
        )

    dataflow_storage_id = df_filt["Dataflow Storage Account ID"].iloc[0]
    payload = {"dataflowStorageId": dataflow_storage_id}

    _base_api(
        request=f"/v1.0/myorg/groups/{workspace_id}/AssignToDataflowStorage",
        method="post",
        payload=payload,
    )

    print(
        f"{icons.green_dot} The '{dataflow_storage_account}' dataflow storage account has been assigned to the '{workspace_name}' workspacce."
    )


def list_dataflow_storage_accounts() -> pd.DataFrame:
    """
    Shows the accessible dataflow storage accounts.

    This is a wrapper function for the following API: `Dataflow Storage Accounts - Get Dataflow Storage Accounts <https://learn.microsoft.com/rest/api/power-bi/dataflow-storage-accounts/get-dataflow-storage-accounts>`_.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the accessible dataflow storage accounts.
    """

    columns = {
        "Dataflow Storage Account ID": "string",
        "Dataflow Storage Account Name": "string",
        "Enabled": "bool",
    }
    df = _create_dataframe(columns=columns)

    response = _base_api(request="/v1.0/myorg/dataflowStorageAccounts")

    for v in response.json().get("value", []):
        new_data = {
            "Dataflow Storage Account ID": v.get("id"),
            "Dataflow Storage Account Name": v.get("name"),
            "Enabled": v.get("isEnabled"),
        }
        df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


def list_upstream_dataflows(
    dataflow: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Shows a list of upstream dataflows for the specified dataflow.

    This is a wrapper function for the following API: `Dataflows - Get Upstream Dataflows In Group <https://learn.microsoft.com/rest/api/power-bi/dataflows/get-upstream-dataflows-in-group>`_.

    Parameters
    ----------
    dataflow : str | uuid.UUID
        Name or UUID of the dataflow.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of upstream dataflows for the specified dataflow.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataflow_name, dataflow_id, dataflow_generation) = (
        _resolve_dataflow_name_and_id_and_generation(
            dataflow=dataflow, workspace=workspace_id
        )
    )

    columns = {
        "Dataflow Name": "string",
        "Dataflow Id": "string",
        "Workspace Name": "string",
        "Workspace Id": "string",
        "Upstream Dataflow Name": "string",
        "Upstream Dataflow Id": "string",
        "Upstream Workspace Name": "string",
        "Upstream Workspace Id": "string",
    }
    df = _create_dataframe(columns=columns)

    def collect_upstreams(dataflow_id, dataflow_name, workspace_id, workspace_name):
        response = _base_api(
            request=f"/v1.0/myorg/groups/{workspace_id}/dataflows/{dataflow_id}/upstreamDataflows"
        )

        values = response.json().get("value", [])
        for v in values:
            tgt_dataflow_id = v.get("targetDataflowId")
            tgt_workspace_id = v.get("groupId")
            tgt_workspace_name = resolve_workspace_name(workspace_id=tgt_workspace_id)
            (tgt_dataflow_name, _, _) = _resolve_dataflow_name_and_id_and_generation(
                dataflow=tgt_dataflow_id, workspace=tgt_workspace_id
            )

            df.loc[len(df)] = {
                "Dataflow Name": dataflow_name,
                "Dataflow Id": dataflow_id,
                "Workspace Name": workspace_name,
                "Workspace Id": workspace_id,
                "Upstream Dataflow Name": tgt_dataflow_name,
                "Upstream Dataflow Id": tgt_dataflow_id,
                "Upstream Workspace Name": tgt_workspace_name,
                "Upstream Workspace Id": tgt_workspace_id,
            }

            collect_upstreams(
                tgt_dataflow_id,
                tgt_dataflow_name,
                tgt_workspace_id,
                tgt_workspace_name,
            )

    collect_upstreams(dataflow_id, dataflow_name, workspace_id, workspace_name)

    return df


def _resolve_dataflow_name_and_id_and_generation(
    dataflow: str | UUID, workspace: Optional[str | UUID] = None
) -> Tuple[str, UUID, str]:

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    dfD = list_dataflows(workspace=workspace_id)

    if _is_valid_uuid(dataflow):
        dfD_filt = dfD[dfD["Dataflow Id"] == dataflow]
    else:
        dfD_filt = dfD[dfD["Dataflow Name"] == dataflow]

    if dfD_filt.empty:
        raise ValueError(
            f"{icons.red_dot} The '{dataflow}' dataflow does not exist within the '{workspace_name}' workspace."
        )

    dataflow_id = dfD_filt["Dataflow Id"].iloc[0]
    dataflow_name = dfD_filt["Dataflow Name"].iloc[0]
    dataflow_generation = dfD_filt["Generation"].iloc[0]

    return (dataflow_name, dataflow_id, dataflow_generation)


def get_dataflow_definition(
    dataflow: str | UUID,
    workspace: Optional[str | UUID] = None,
    decode: bool = True,
) -> dict:
    """
    Obtains the definition of a dataflow. This supports Gen1, Gen2 and Gen 2 CI/CD dataflows.

    This is a wrapper function for the following API: `Dataflows - Get Dataflow <https://learn.microsoft.com/rest/api/power-bi/dataflows/get-dataflow>`_.

    Parameters
    ----------
    dataflow : str | uuid.UUID
        The name or ID of the dataflow.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name.
        Defaults to None, which resolves to the workspace of the attached lakehouse
        or if no lakehouse is attached, resolves to the workspace of the notebook.
    decode : bool, optional
        If True, decodes the dataflow definition file.

    Returns
    -------
    dict
        The dataflow definition.
    """

    workspace_id = resolve_workspace_id(workspace)

    (dataflow_name, dataflow_id, dataflow_generation) = (
        _resolve_dataflow_name_and_id_and_generation(
            dataflow=dataflow, workspace=workspace_id
        )
    )

    if dataflow_generation == "Gen2 CI/CD":
        result = _base_api(
            request=f"/v1/workspaces/{workspace_id}/items/{dataflow_id}/getDefinition",
            client="fabric_sp",
            method="post",
            lro_return_json=True,
            status_codes=[200, 202],
        )

        if decode:
            # Decode the payload from base64
            definition = {"definition": {"parts": []}}

            for part in result.get("definition", {}).get("parts", []):
                path = part.get("path")
                payload = part.get("payload")
                decoded_payload = _decode_b64(payload)
                definition["definition"]["parts"].append(
                    {"path": path, "payload": decoded_payload}
                )
            return definition
        else:
            return result
    else:
        result = _base_api(
            request=f"/v1.0/myorg/groups/{workspace_id}/dataflows/{dataflow_id}",
            client="fabric_sp",
        ).json()

        return result


def upgrade_dataflow(
    dataflow: str | UUID,
    workspace: Optional[str | UUID] = None,
    new_dataflow_name: Optional[str] = None,
    new_dataflow_workspace: Optional[str | UUID] = None,
):
    """
    Creates a Dataflow Gen2 CI/CD item based on the mashup definition from an existing Gen1/Gen2 dataflow. After running this function, update the connections in the dataflow to ensure the data can be properly refreshed.

    Parameters
    ----------
    dataflow : str | uuid.UUID
        The name or ID of the dataflow.
    workspace : str | uuid.UUID, default=None
        The workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    new_dataflow_name: str, default=None
        Name of the new dataflow.
    new_dataflow_workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID of the dataflow to be created.
        Defaults to None which resolves to the existing workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    # Resolve the workspace name and ID
    (workspace, workspace_id) = resolve_workspace_name_and_id(workspace)

    # Resolve the dataflow name and ID
    (dataflow_name, dataflow_id, dataflow_generation) = (
        _resolve_dataflow_name_and_id_and_generation(dataflow, workspace_id)
    )

    if dataflow_generation == "Gen2 CI/CD":
        # Print an error message that the dataflow is already a native Fabric item
        print(
            f"{icons.info} The dataflow '{dataflow_name}' is already a Fabric native Dataflow Gen2 item. No changes made."
        )
        return

    (new_dataflow_workspace, new_dataflow_workspace_id) = resolve_workspace_name_and_id(
        new_dataflow_workspace
    )

    # If no new dataflow name is provided, use the existing dataflow name
    if not new_dataflow_name:
        new_dataflow_name = dataflow_name

    # Get dataflow definition
    definition = get_dataflow_definition(dataflow, workspace_id)

    # Check for linked table references
    matches = (
        parse("$['pbi:mashup'].connectionOverrides[*].kind").find(definition) or []
    )
    if any(match.value in {"PowerPlatformDataflows", "PowerBI"} for match in matches):
        print(
            f"""{icons.red_dot} The dataflow '{dataflow_name}' contains a linked table reference to an existing dataflow as a connection source and will not be upgraded. No changes were made.
    - To track the upstream lineage of linked tables across dataflows use the list_upstream_dataflows function.
    - To automatically remove the tables and upgrade the existing dataflow use the upgrade_powerbippdf_dataflow function."""
        )
        return

    description = get_jsonpath_value(data=definition, path="$.description")

    payload = {
        "displayName": new_dataflow_name,
    }
    if description:
        payload["description"] = description

    # Query Groups
    matches = parse("$.annotations[?(@.name=='pbi:QueryGroups')].value").find(
        definition
    )
    query_groups_value = json.loads(matches[0].value) if matches else []

    queries_metadata = get_jsonpath_value(
        data=definition, path="$['pbi:mashup'].queriesMetadata"
    )

    default_staging = True if "DefaultStaging" in queries_metadata else False

    # Collect keys to delete
    keys_to_delete = [
        key
        for key in queries_metadata
        if key.endswith("_DataDestination")
        or key.endswith("_WriteToDataDestination")
        or key.endswith("_TransformForWriteToDataDestination")
        or key == "FastCopyStaging"
    ]

    # Delete them
    for key in keys_to_delete:
        del queries_metadata[key]

    # Set load enabled and isHidden
    for key, items in queries_metadata.items():
        items["loadEnabled"] = False
        if key in ["DefaultDestination", "DefaultStaging"]:
            items["isHidden"] = True

    # Prepare the dataflow definition
    query_metadata = {
        "formatVersion": "202502",
        "computeEngineSettings": {},  # How to set this?
        "name": new_dataflow_name,
        "queryGroups": query_groups_value,
        "documentLocale": get_jsonpath_value(data=definition, path="$.culture"),
        "queriesMetadata": queries_metadata,
        "fastCombine": get_jsonpath_value(
            data=definition, path="$['pbi:mashup'].fastCombine", default=False
        ),
        "allowNativeQueries": get_jsonpath_value(
            data=definition, path="$['pbi:mashup'].allowNativeQueries", default=False
        ),
        # "connections": [],
    }

    fast_copy = get_jsonpath_value(
        data=definition, path="$['ppdf:fastCopy']", default=False
    )
    max_concurrency = get_jsonpath_value(
        data=definition, path="$['ppdf:maxConcurrency']"
    )
    if fast_copy:
        query_metadata["computeEngineSettings"] = {}

        if max_concurrency:
            query_metadata["computeEngineSettings"]["maxConcurrency"] = max_concurrency

    mashup_doc = get_jsonpath_value(data=definition, path="$['pbi:mashup'].document")

    # Remove the FastCopyStaging section if it exists
    new_mashup_doc = ""
    if default_staging and fast_copy:
        new_mashup_doc = '[DefaultOutputDestinationSettings = [DestinationDefinition = [Kind = "Reference", QueryName = "DefaultDestination", IsNewTarget = true], UpdateMethod = [Kind = "Replace"]], StagingDefinition = [Kind = "FastCopy"]]\r\nsection Section1'
    elif default_staging and not fast_copy:
        new_mashup_doc = '[DefaultOutputDestinationSettings = [DestinationDefinition = [Kind = "Reference", QueryName = "DefaultDestination", IsNewTarget = true], UpdateMethod = [Kind = "Replace"]]\r\nsection Section1'
    elif not default_staging and fast_copy:
        new_mashup_doc = '[StagingDefinition = [Kind = "FastCopy"]]\r\nsection Section1'
    else:
        new_mashup_doc = "section Section1"
    for i in mashup_doc.split(";\r\nshared "):
        # if 'IsParameterQuery=true' in i:
        # Add to queries_metadata
        if not (
            "FastCopyStaging = let" in i
            or '_WriteToDataDestination" = let' in i
            or "_WriteToDataDestination = let" in i
            or '_DataDestination" = let' in i
            or "_DataDestination = let" in i
            or '_TransformForWriteToDataDestination" = let' in i
            or "_TransformForWriteToDataDestination = let" in i
        ):
            if i != "section Section1":
                if default_staging and (
                    "IsParameterQuery=true" not in i
                    and not i.startswith("DefaultStaging")
                    and not i.startswith("DefaultDestination")
                ):
                    new_mashup_doc += (
                        ";\r\n[BindToDefaultDestination = true]\r\nshared " + i
                    )
                else:
                    new_mashup_doc += ";\r\nshared " + i
    new_mashup_doc = f"{new_mashup_doc};"

    return new_mashup_doc, query_metadata

    # Add the dataflow definition to the payload
    new_definition = {
        "parts": [
            {
                "path": "queryMetadata.json",
                "payload": _conv_b64(query_metadata),
                "payloadType": "InlineBase64",
            },
            {
                "path": "mashup.pq",
                "payload": _conv_b64(new_mashup_doc, json_dumps=False),
                "payloadType": "InlineBase64",
            },
        ]
    }

    create_dataflow(
        name=new_dataflow_name,
        workspace=new_dataflow_workspace,
        definition=new_definition,
    )


def create_dataflow(
    name: str,
    workspace: Optional[str | UUID] = None,
    description: Optional[str] = None,
    definition: Optional[dict] = None,
):
    """
    Creates a native Fabric Dataflow Gen2 CI/CD item.

    Parameters
    ----------
    name : str
        The name the dataflow.
    workspace : str | uuid.UUID, default=None
        The workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    description : str, default=None
        The description of the dataflow.
    definition : dict, default=None
        The definition of the dataflow in the form of a dictionary.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    payload = {
        "displayName": name,
    }
    if description:
        payload["description"] = description

    if definition:
        payload["definition"] = definition

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/dataflows",
        method="post",
        payload=payload,
        client="fabric_sp",
        lro_return_json=True,
        status_codes=[201, 202],
    )

    print(
        f"{icons.green_dot} The dataflow '{name}' has been created within the '{workspace_name}' workspace."
    )



from sempy._utils._log import log
from sempy_labs._helper_functions import (
    _base_api,
    _is_valid_uuid,
)
import sempy_labs._icons as icons
from typing import List


@log
def remove_all_sharing_links(sharing_link_type: str = "OrgLink"):
    """
    Deletes all organization sharing links for all Fabric items in the tenant. This action cannot be undone.

    This is a wrapper function for the following API: `Sharing Links - Remove All Sharing Links <https://learn.microsoft.com/rest/api/fabric/admin/sharing-links/remove-all-sharing-links>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    sharing_link_type : str, default='OrgLink'
        Specifies the type of sharing link that is required to be deleted. Additional sharing link types may be added over time.
    """

    payload = {"sharingLinkType": sharing_link_type}

    _base_api(
        request="/v1/admin/items/removeAllSharingLinks",
        client="fabric_sp",
        lro_return_status_code=True,
        status_codes=[200, 202],
        method="post",
        payload=payload,
    )

    print(
        f"{icons.green_dot} All organization sharing links for all Fbric items in the tenant have been deleted."
    )


@log
def remove_sharing_links(items: List[dict], sharing_link_type: str = "OrgLink"):
    """
    Deletes all organization sharing links for the specified Fabric items. This action cannot be undone.

    This is a wrapper function for the following API: `Sharing Links - Bulk Remove Sharing Links <https://learn.microsoft.com/rest/api/fabric/admin/sharing-links/bulk-remove-sharing-links>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    items : List[dict]
        A list of dictionaries, each representing an item. The 'item' and 'workspace' accepts either name or ID. Examples:

        [
            {
                "item": "MyReport",
                "type": "Report",
                "workspace": "Workpsace 1",
            },
            {
                "item": "MyReport2",
                "type": "Report",
                "workspace": "Workspace 2",
            },
        ]

        [
            {
                "item": "fe472f5e-636e-4c10-a1c6-7e9edc0b542a",
                "type": "Report",
                "workspace": "Workspace 1",
            },
            {
                "item": "fe472f5e-636e-4c10-a1c6-7e9edc0b542c",
                "type": "Notebook",
                "workspace": "476fcafe-b514-495d-b13f-ca9a4f0b1d8f",
            },
        ]

    sharing_link_type : str, default='OrgLink'
        Specifies the type of sharing link that is required to be deleted. Additional sharing link types may be added over time.
    """

    from sempy_labs.admin._items import _resolve_item_id

    payload = {"items": [], "sharingLinkType": sharing_link_type}

    for i in items:
        item = i.get("item")
        type = i.get("type")
        workspace = i.get("workspace")
        if _is_valid_uuid(item):
            payload["items"].append({"id": item, "type": type})
        else:
            item_id = _resolve_item_id(item=item, type=type, workspace=workspace)
            payload["items"].append({"id": item_id, "type": type})

    _base_api(
        request="/v1/admin/items/bulkRemoveSharingLinks",
        client="fabric_sp",
        method="post",
        payload=payload,
        lro_return_status_code=True,
        status_codes=[200, 202],
    )

    print(
        f"{icons.green_dot} Organizational sharing links for all specified items have been deleted."
    )



import sempy
from ..tom import connect_semantic_model
from .._refresh_semantic_model import refresh_semantic_model
from ._dl_helper import get_direct_lake_source
from .._helper_functions import (
    _convert_data_type,
    resolve_dataset_name_and_id,
    resolve_workspace_name_and_id,
    resolve_workspace_name,
)
from sempy._utils._log import log
from typing import List, Optional, Union
import sempy_labs._icons as icons
from uuid import UUID


@log
def update_direct_lake_partition_entity(
    dataset: str | UUID,
    table_name: Union[str, List[str]],
    entity_name: Union[str, List[str]],
    schema: Optional[str] = None,
    workspace: Optional[str | UUID] = None,
):
    """
    Remaps a table (or tables) in a Direct Lake semantic model to a table in a lakehouse.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    table_name : str, List[str]
        Name of the table(s) in the semantic model.
    entity_name : str, List[str]
        Name of the lakehouse table to be mapped to the semantic model table.
    schema : str, default=None
        The schema of the lakehouse table to be mapped to the semantic model table.
        Defaults to None which resolves to the existing schema of the lakehouse table.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID in which the semantic model exists.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    # Support both str & list types
    if isinstance(table_name, str):
        table_name = [table_name]
    if isinstance(entity_name, str):
        entity_name = [entity_name]

    if len(table_name) != len(entity_name):
        raise ValueError(
            f"{icons.red_dot} The 'table_name' and 'entity_name' arrays must be of equal length."
        )

    icons.sll_tags.append("UpdateDLPartition")

    with connect_semantic_model(
        dataset=dataset_id, readonly=False, workspace=workspace_id
    ) as tom:

        if not tom.is_direct_lake():
            raise ValueError(
                f"{icons.red_dot} The '{dataset_name}' semantic model within the '{workspace_name}' workspace is not in Direct Lake mode."
            )

        for tName in table_name:
            i = table_name.index(tName)
            eName = entity_name[i]
            part_name = next(
                p.Name
                for t in tom.model.Tables
                for p in t.Partitions
                if t.Name == tName
            )

            if part_name is None:
                raise ValueError(
                    f"{icons.red_dot} The '{tName}' table in the '{dataset_name}' semantic model has not been updated."
                )

            tom.model.Tables[tName].Partitions[part_name].Source.EntityName = eName

            # Update source lineage tag
            existing_schema = (
                tom.model.Tables[tName].Partitions[part_name].Source.SchemaName or "dbo"
            )
            if schema is None:
                schema = existing_schema

            tom.model.Tables[tName].Partitions[part_name].Source.SchemaName = schema
            tom.model.Tables[tName].SourceLineageTag = f"[{schema}].[{eName}]"
            print(
                f"{icons.green_dot} The '{tName}' table in the '{dataset_name}' semantic model within the '{workspace_name}' workspace has been updated to point to the '{eName}' table."
            )


@log
def add_table_to_direct_lake_semantic_model(
    dataset: str | UUID,
    table_name: str,
    lakehouse_table_name: str,
    refresh: bool = True,
    workspace: Optional[str | UUID] = None,
):
    """
    Adds a table and all of its columns to a Direct Lake semantic model, based on a Fabric lakehouse table.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    table_name : str, List[str]
        Name of the table in the semantic model.
    lakehouse_table_name : str
        The name of the Fabric lakehouse table.
    refresh : bool, default=True
        Refreshes the table after it is added to the semantic model.
    workspace : str | uuid.UUID, default=None
        The name or ID of the Fabric workspace in which the semantic model resides.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    sempy.fabric._client._utils._init_analysis_services()
    import Microsoft.AnalysisServices.Tabular as TOM
    from sempy_labs.lakehouse._get_lakehouse_columns import get_lakehouse_columns
    from sempy_labs.lakehouse._get_lakehouse_tables import get_lakehouse_tables

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    artifact_type, lakehouse_name, lakehouse_id, lakehouse_workspace_id = (
        get_direct_lake_source(dataset=dataset_id, workspace=workspace_id)
    )

    if artifact_type == "Warehouse":
        raise ValueError(
            f"{icons.red_dot} This function is only valid for Direct Lake semantic models which source from Fabric lakehouses (not warehouses)."
        )

    if artifact_type is None:
        raise ValueError(
            f"{icons.red_dot} This function only supports Direct Lake semantic models where the source lakehouse resides in the same workpace as the semantic model."
        )

    lakehouse_workspace = resolve_workspace_name(workspace_id=lakehouse_workspace_id)

    with connect_semantic_model(
        dataset=dataset_id, readonly=False, workspace=workspace_id
    ) as tom:

        table_count = tom.model.Tables.Count

        if not tom.is_direct_lake() and table_count > 0:
            raise ValueError(
                f"{icons.red_dot} This function is only valid for Direct Lake semantic models or semantic models with no tables."
            )

        if any(
            p.Name == lakehouse_table_name
            for p in tom.all_partitions()
            if p.SourceType == TOM.PartitionSourceType.Entity
        ):
            t_name = next(
                p.Parent.Name
                for p in tom.all_partitions()
                if p.Name
                == lakehouse_table_name & p.SourceType
                == TOM.PartitionSourceType.Entity
            )
            raise ValueError(
                f"The '{lakehouse_table_name}' table already exists in the '{dataset_name}' semantic model within the '{workspace_name}' workspace as the '{t_name}' table."
            )

        if any(t.Name == table_name for t in tom.model.Tables):
            raise ValueError(
                f"The '{table_name}' table already exists in the '{dataset_name}' semantic model within the '{workspace_name}' workspace."
            )

        dfL = get_lakehouse_tables(
            lakehouse=lakehouse_name, workspace=lakehouse_workspace
        )
        dfL_filt = dfL[dfL["Table Name"] == lakehouse_table_name]

        if len(dfL_filt) == 0:
            raise ValueError(
                f"The '{lakehouse_table_name}' table does not exist in the '{lakehouse_name}' lakehouse within the '{lakehouse_workspace}' workspace."
            )

        dfLC = get_lakehouse_columns(
            lakehouse=lakehouse_name, workspace=lakehouse_workspace
        )
        dfLC_filt = dfLC[dfLC["Table Name"] == lakehouse_table_name]

        tom.add_table(name=table_name)
        print(
            f"{icons.green_dot} The '{table_name}' table has been added to the '{dataset_name}' semantic model within the '{workspace_name}' workspace."
        )
        tom.add_entity_partition(
            table_name=table_name, entity_name=lakehouse_table_name
        )
        print(
            f"{icons.green_dot} The '{lakehouse_table_name}' partition has been added to the '{table_name}' table in the '{dataset_name}' semantic model within the '{workspace_name}' workspace."
        )

        for i, r in dfLC_filt.iterrows():
            lakeCName = r["Column Name"]
            dType = r["Data Type"]
            dt = _convert_data_type(dType)
            tom.add_data_column(
                table_name=table_name,
                column_name=lakeCName,
                source_column=lakeCName,
                data_type=dt,
            )
            print(
                f"{icons.green_dot} The '{lakeCName}' column has been added to the '{table_name}' table as a '{dt}' data type in the '{dataset_name}' semantic model within the '{workspace_name}' workspace."
            )

        if refresh:
            refresh_semantic_model(
                dataset=dataset_id, tables=table_name, workspace=workspace_id
            )



import pandas as pd
import sempy_labs._icons as icons
from typing import Optional, List
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    _base_api,
    _create_dataframe,
)
from uuid import UUID
from sempy._utils._log import log


@log
def connect_workspace_to_azure_dev_ops(
    organization_name: str,
    project_name: str,
    repository_name: str,
    branch_name: str,
    directory_name: str,
    workspace: Optional[str | UUID] = None,
):
    """
    Connects a workspace to an Azure DevOps git repository.

    This is a wrapper function for the following API: `Git - Connect <https://learn.microsoft.com/rest/api/fabric/core/git/connect>`_.

    Parameters
    ----------
    organization_name : str
        The organization name.
    project_name : str
        The project name.
    repository_name : str
        The repository name.
    branch_name : str
        The branch name.
    directory_name : str
        The directory name.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    payload = {
        "gitProviderDetails": {
            "organizationName": organization_name,
            "projectName": project_name,
            "gitProviderType": "AzureDevOps",
            "repositoryName": repository_name,
            "branchName": branch_name,
            "directoryName": directory_name,
        }
    }

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/git/connect",
        payload=payload,
        method="post",
    )

    print(
        f"{icons.green_dot} The '{workspace_name}' workspace has been connected to the '{project_name}' Git project in Azure DevOps within the '{repository_name}' repository."
    )


@log
def connect_workspace_to_github(
    owner_name: str,
    repository_name: str,
    branch_name: str,
    directory_name: str,
    connection_id: UUID,
    source: str = "ConfiguredConnection",
    workspace: Optional[str | UUID] = None,
):
    """
    Connects a workspace to a GitHub git repository.

    This is a wrapper function for the following API: `Git - Connect <https://learn.microsoft.com/rest/api/fabric/core/git/connect>`_.

    Parameters
    ----------
    owner_name : str
        The owner name.
    repository_name : str
        The repository name.
    branch_name : str
        The branch name.
    directory_name : str
        The directory name.
    source : str, default="ConfiguredConnection"
        The Git credentials source.
    connection_id : uuid.UUID
        The object ID of the connection.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    payload = {
        "gitProviderDetails": {
            "ownerName": owner_name,
            "gitProviderType": "GitHub",
            "repositoryName": repository_name,
            "branchName": branch_name,
            "directoryName": directory_name,
        },
        "myGitCredentials": {
            "source": source,
            "connectionId": connection_id,
        },
    }

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/git/connect",
        payload=payload,
        method="post",
    )

    print(
        f"{icons.green_dot} The '{workspace_name}' workspace has been connected to the '{repository_name}' GitHub repository."
    )


@log
def disconnect_workspace_from_git(workspace: Optional[str | UUID] = None):
    """
    Disconnects a workspace from a git repository.

    This is a wrapper function for the following API: `Git - Disconnect <https://learn.microsoft.com/rest/api/fabric/core/git/disconnect>`_.

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    _base_api(request=f"/v1/workspaces/{workspace_id}/git/disconnect", method="post")

    print(
        f"{icons.green_dot} The '{workspace_name}' workspace has been disconnected from Git."
    )


@log
def get_git_status(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Obtains the Git status of items in the workspace, that can be committed to Git.

    This is a wrapper function for the following API: `Git - Get Status <https://learn.microsoft.com/rest/api/fabric/core/git/get-status>`_.

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the Git status of items in the workspace.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    df = pd.DataFrame(
        columns=[
            "Workspace Head",
            "Remote Commit Hash",
            "Object ID",
            "Logical ID",
            "Item Type",
            "Item Name",
            "Workspace Change",
            "Remote Change",
            "Conflict Type",
        ]
    )

    result = _base_api(
        request=f"/v1/workspaces/{workspace_id}/git/status",
        lro_return_json=True,
        status_codes=None,
    )

    for changes in result.get("changes", []):
        item_metadata = changes.get("itemMetadata", {})
        item_identifier = item_metadata.get("itemIdentifier", {})

        new_data = {
            "Workspace Head": result.get("workspaceHead"),
            "Remote Commit Hash": result.get("remoteCommitHash"),
            "Object ID": item_identifier.get("objectId"),
            "Logical ID": item_identifier.get("logicalId"),
            "Item Type": item_metadata.get("itemType"),
            "Item Name": item_metadata.get("displayName"),
            "Remote Change": changes.get("remoteChange"),
            "Workspace Change": changes.get("workspaceChange"),
            "Conflict Type": changes.get("conflictType"),
        }
        df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    return df


@log
def get_git_connection(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Obtains the Git status of items in the workspace, that can be committed to Git.

    This is a wrapper function for the following API: `Git - Get Connection <https://learn.microsoft.com/rest/api/fabric/core/git/get-connection>`_.

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the Git status of items in the workspace.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    df = pd.DataFrame(
        columns=[
            "Organization Name",
            "Project Name",
            "Git Provider Type",
            "Repository Name",
            "Branch Name",
            "Directory Name",
            "Workspace Head",
            "Last Sync Time",
            "Git Connection State",
        ]
    )

    response = _base_api(request=f"/v1/workspaces/{workspace_id}/git/connection")

    r = response.json()
    provider_details = r.get("gitProviderDetails", {})
    sync_details = r.get("gitSyncDetails", {})
    new_data = {
        "Organization Name": provider_details.get("organizationName"),
        "Project Name": provider_details.get("projectName"),
        "Git Provider Type": provider_details.get("gitProviderType"),
        "Repository Name": provider_details.get("repositoryName"),
        "Branch Name": provider_details.get("branchName"),
        "Directory Name": provider_details.get("directoryName"),
        "Workspace Head": sync_details.get("head"),
        "Last Sync Time": sync_details.get("lastSyncTime"),
        "Git Connection State": r.get("gitConnectionState"),
    }
    df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    return df


@log
def initialize_git_connection(workspace: Optional[str | UUID] = None) -> str:
    """
    Initializes a connection for a workspace that is connected to Git.

    This is a wrapper function for the following API: `Git - Initialize Connection <https://learn.microsoft.com/rest/api/fabric/core/git/initialize-connection>`_.

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    str
        Remote full SHA commit hash.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    response_json = _base_api(
        request=f"/v1/workspaces/{workspace_id}/git/initializeConnection",
        method="post",
        lro_return_json=True,
        status_codes=None,
    )

    print(
        f"{icons.green_dot} The '{workspace_name}' workspace git connection has been initialized."
    )

    return response_json.get("remoteCommitHash")


@log
def commit_to_git(
    comment: str,
    item_ids: UUID | List[UUID] = None,
    workspace: Optional[str | UUID] = None,
):
    """
    Commits all or a selection of items within a workspace to Git.

    This is a wrapper function for the following API: `Git - Commit to Git <https://learn.microsoft.com/rest/api/fabric/core/git/commit-to-git>`_.

    Parameters
    ----------
    comment : str
        The Git commit comment.
    item_ids : uuid.UUID | List[uuid.UUID], default=None
        A list of item Ids to commit to Git.
        Defaults to None which commits all items to Git.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    gs = get_git_status(workspace=workspace_id)
    if not gs.empty:
        workspace_head = gs["Workspace Head"].iloc[0]

        if item_ids is None:
            commit_mode = "All"
        else:
            commit_mode = "Selective"

        if isinstance(item_ids, str):
            item_ids = [item_ids]

        payload = {
            "mode": commit_mode,
            "workspaceHead": workspace_head,
            "comment": comment,
        }

        if item_ids is not None:
            payload["items"] = [{"objectId": item_id} for item_id in item_ids]

        _base_api(
            request=f"/v1/workspaces/{workspace_id}/git/commitToGit",
            method="post",
            payload=payload,
            lro_return_status_code=True,
            status_codes=None,
        )

        if commit_mode == "All":
            print(
                f"{icons.green_dot} All items within the '{workspace_name}' workspace have been committed to Git."
            )
        else:
            print(
                f"{icons.green_dot} The {item_ids} items within the '{workspace_name}' workspace have been committed to Git."
            )
    else:
        print(
            f"{icons.info} Git already up to date: no modified items found within the '{workspace_name}' workspace."
        )


@log
def update_from_git(
    remote_commit_hash: str,
    conflict_resolution_policy: str,
    workspace_head: Optional[str] = None,
    allow_override: bool = False,
    workspace: Optional[str | UUID] = None,
):
    """
    Updates the workspace with commits pushed to the connected branch.

    This is a wrapper function for the following API: `Git - Update From Git <https://learn.microsoft.com/rest/api/fabric/core/git/update-from-git>`_.

    Parameters
    ----------
    remote_commit_hash : str
        Remote full SHA commit hash.
    confilict_resolution_policy : str
        The `conflict resolution policy <https://learn.microsoft.com/rest/api/fabric/core/git/update-from-git?tabs=HTTP#conflictresolutionpolicy>`_.
    workspace_head : str
        Full SHA hash that the workspace is synced to. This value may be null only after Initialize Connection.
        In other cases, the system will validate that the given value is aligned with the head known to the system.
    allow_override : bool, default=False
        User consent to override incoming items during the update from Git process. When incoming items are present and the allow override items is not specified or is provided as false, the update operation will not start. Default value is false.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    conflict_resolution_policies = ["PreferWorkspace", "PreferRemote"]
    if "remote" in [policy.lower() for policy in conflict_resolution_policies]:
        conflict_resolution_policies = "PreferRemote"
    elif "workspace" in [policy.lower() for policy in conflict_resolution_policies]:
        conflict_resolution_policies = "PreferWorkspace"

    if conflict_resolution_policy not in conflict_resolution_policies:
        raise ValueError(
            f"{icons.red_dot} Invalid conflict resolution policy. Valid options: {conflict_resolution_policies}."
        )

    payload = {}
    payload["remoteCommitHash"] = remote_commit_hash
    payload["conflictResolution"] = {
        "conflictResolutionType": "Workspace",
        "conflictResolutionPolicy": conflict_resolution_policy,
    }

    if workspace_head is not None:
        payload["workspaceHead"] = workspace_head
    if allow_override is not None:
        payload["options"] = {"allowOverrideItems": allow_override}

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/git/updateFromGit",
        method="post",
        payload=payload,
        lro_return_status_code=True,
        status_codes=None,
    )

    print(
        f"{icons.green_dot} The '{workspace_name}' workspace has been updated with commits pushed to the connected branch."
    )


@log
def get_my_git_credentials(
    workspace: Optional[str | UUID] = None,
) -> pd.DataFrame:
    """
    Returns the user's Git credentials configuration details.

    This is a wrapper function for the following API: `Git - Get My Git Credentials <https://learn.microsoft.com/rest/api/fabric/core/git/get-my-git-credentials>`_.

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the user's Git credentials configuration details.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    columns = {
        "Source": "string",
    }

    df = _create_dataframe(columns)

    response = _base_api(request=f"/v1/workspaces/{workspace_id}/git/myGitCredentials")

    r = response.json()
    new_data = {
        "Source": r.get("source"),
        "Connection Id": r.get("connectionId"),
    }
    df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    return df


@log
def update_my_git_credentials(
    source: str,
    connection_id: Optional[UUID] = None,
    workspace: Optional[str | UUID] = None,
):
    """
    Updates the user's Git credentials configuration details.

    This is a wrapper function for the following API: `Git - Update My Git Credentials <https://learn.microsoft.com/rest/api/fabric/core/git/update-my-git-credentials>`_.

    Parameters
    ----------
    source : str
        The Git credentials source. Valid options: 'Automatic', 'ConfiguredConnection', 'None'.
    connection_id : UUID, default=None
        The object ID of the connection. Valid only for the 'ConfiguredConnection' source.
    workspace : str | uuid.UUID, default=None
        The workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    if source == "ConfiguredConnection" and connection_id is None:
        raise ValueError(
            f"{icons.red_dot} The 'ConfiguredConnection' source requires a connection_id."
        )

    payload = {
        "source": source,
    }

    if connection_id is not None:
        payload["connectionId"] = connection_id

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/git/myGitCredentials",
        method="patch",
        payload=payload,
    )

    print(
        f"{icons.green_dot} The user's Git credentials have been updated accordingly."
    )



from typing import Optional
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    resolve_lakehouse_name_and_id,
    _base_api,
)
from uuid import UUID
from sempy._utils._log import log
import sempy_labs._icons as icons
import time
from sempy_labs._job_scheduler import (
    _get_item_job_instance,
)
import pandas as pd


@log
def refresh_materialized_lake_views(
    lakehouse: Optional[str | UUID] = None, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Run on-demand Refresh MaterializedLakeViews job instance.

    This is a wrapper function for the following API: `Background Jobs - Run On Demand Refresh Materialized Lake Views <https://learn.microsoft.com/rest/api/fabric/lakehouse/background-jobs/run-on-demand-refresh-materialized-lake-views>`_.

    Parameters
    ----------
    lakehouse : str | uuid.UUID, default=None
        The Fabric lakehouse name or ID.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID used by the lakehouse.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A DataFrame containing the job instance details of the refresh materialized lake views operation.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (lakehouse_name, lakehouse_id) = resolve_lakehouse_name_and_id(
        lakehouse=lakehouse, workspace=workspace_id
    )

    response = _base_api(
        request=f"/v1/workspaces/{workspace_id}/lakehouses/{lakehouse_id}/jobs/instances?jobType=RefreshMaterializedLakeViews",
        method="post",
        status_codes=[202],
    )

    print(
        f"{icons.in_progress} The refresh materialized lake views job for the '{lakehouse_name}' lakehouse within the '{workspace_name}' workspace has been initiated."
    )

    status_url = response.headers.get("Location").split("fabric.microsoft.com")[1]
    status = None
    while status not in ["Completed", "Failed"]:
        response = _base_api(request=status_url)
        status = response.json().get("status")
        time.sleep(3)

    df = _get_item_job_instance(url=status_url)

    if status == "Completed":
        print(
            f"{icons.green_dot} The refresh materialized lake views job for the '{lakehouse_name}' lakehouse within the '{workspace_name}' workspace has succeeded."
        )
    else:
        print(status)
        print(
            f"{icons.red_dot} The refresh materialized lake views job for the '{lakehouse_name}' lakehouse within the '{workspace_name}' workspace has failed."
        )

    return df



import pandas as pd
from uuid import UUID
from sempy_labs._helper_functions import (
    _is_valid_uuid,
    _base_api,
    _create_dataframe,
    _update_dataframe_datatypes,
)
from sempy._utils._log import log
import sempy_labs._icons as icons
from typing import List, Literal, Optional


@log
def resolve_group_id(group: str | UUID) -> UUID:
    """
    Resolves the group ID from the group name or ID.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    group : str | uuid.UUID
        The group name or ID.

    Returns
    -------
    uuid.UUID
        The group ID.
    """
    if _is_valid_uuid(group):
        group_id = group
    else:
        dfG = list_groups()
        dfG_filt = dfG[dfG["Group Name"] == group]
        if dfG_filt.empty:
            raise ValueError(f"{icons.red_dot} The '{group}' group does not exist.")
        group_id = dfG_filt["Group Id"].iloc[0]

    return group_id


@log
def list_groups() -> pd.DataFrame:
    """
    Shows a list of groups and their properties.

    This is a wrapper function for the following API: `List groups <https://learn.microsoft.com/graph/api/group-list>`_.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of groups and their properties.
    """

    result = _base_api(request="groups", client="graph", uses_pagination=True)

    columns = {
        "Group Id": "string",
        "Group Name": "string",
        "Mail": "string",
        "Description": "string",
        "Classification": "string",
        "Mail Enabled": "bool",
        "Security Enabled": "bool",
        "Created Date Time": "datetime",
        "Expiration Date Time": "string",
        "Deleted Date Time": "string",
        "Renewed Date Time": "string",
        "Visibility": "string",
        "Security Identifier": "string",
    }

    df = _create_dataframe(columns=columns)

    rows = []
    for r in result:
        for v in r.get("value", []):
            rows.append(
                {
                    "Group Id": v.get("id"),
                    "Group Name": v.get("displayName"),
                    "Mail": v.get("mail"),
                    "Description": v.get("description"),
                    "Classification": v.get("classification"),
                    "Mail Enabled": v.get("mailEnabled"),
                    "Security Enabled": v.get("securityEnabled"),
                    "Created Date Time": v.get("createdDateTime"),
                    "Expiration Date Time": v.get("expirationDateTime"),
                    "Renewed Date Time": v.get("renewedDateTime"),
                    "Deleted Date Time": v.get("deletedDateTime"),
                    "Visibility": v.get("visibility"),
                    "Security Identifier": v.get("securityIdentifier"),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


@log
def _get_group(group_id: UUID) -> pd.DataFrame:
    """
    Shows a list of groups and their properties.

    This is a wrapper function for the following API: `Get group <https://learn.microsoft.com/graph/api/group-get>`_.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    group_id : uuid.UUID
        The group ID.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of groups and their properties.
    """

    result = _base_api(request=f"groups/{group_id}", client="graph").json()

    columns = {
        "Group Id": "string",
        "Group Name": "string",
        "Mail": "string",
        "Description": "string",
        "Classification": "string",
        "Mail Enabled": "bool",
        "Security Enabled": "bool",
        "Created Date Time": "datetime",
        "Expiration Date Time": "string",
        "Deleted Date Time": "string",
        "Renewed Date Time": "string",
        "Visibility": "string",
        "Security Identifier": "string",
    }
    df = _create_dataframe(columns=columns)

    rows = []
    for v in result.get("value"):
        rows.append(
            {
                "Group Id": v.get("id"),
                "Group Name": v.get("displayName"),
                "Mail": v.get("mail"),
                "Description": v.get("description"),
                "Classification": v.get("classification"),
                "Mail Enabled": v.get("mailEnabled"),
                "Security Enabled": v.get("securityEnabled"),
                "Created Date Time": v.get("createdDateTime"),
                "Expiration Date Time": v.get("expirationDateTime"),
                "Deleted Date Time": v.get("deletedDateTime"),
                "Renewed Date Time": v.get("renewedDateTime"),
                "Visibility": v.get("visibility"),
                "Security Identifier": v.get("securityIdentifier"),
            }
        )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


@log
def list_group_members(group: str | UUID) -> pd.DataFrame:
    """
    Shows a list of the members of a group.

    This is a wrapper function for the following API: `List group members <https://learn.microsoft.com/graph/api/group-list-members>`_.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    group : str | uuid.UUID
        The group name or ID.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of the members of a group.
    """

    group_id = resolve_group_id(group)

    result = _base_api(
        request=f"groups/{group_id}/members", client="graph", uses_pagination=True
    )

    columns = {
        "Member Id": "string",
        "Member Name": "string",
        "User Principal Name": "string",
        "Mail": "string",
        "Job Title": "string",
        "Office Location": "string",
        "Mobile Phone": "string",
        "Business Phones": "string",
        "Preferred Language": "string",
        "Given Name": "string",
        "Surname": "string",
    }

    df = _create_dataframe(columns=columns)

    rows = []
    for r in result:
        for v in r.get("value", []):
            rows.append(
                {
                    "Member Id": v.get("id"),
                    "Member Name": v.get("displayName"),
                    "User Principal Name": v.get("userPrincipalName"),
                    "Mail": v.get("mail"),
                    "Job Title": v.get("jobTitle"),
                    "Office Location": v.get("officeLocation"),
                    "Mobile Phone": v.get("mobilePhone"),
                    "Business Phones": str(v.get("businessPhones")),
                    "Preferred Language": v.get("preferredLanguage"),
                    "Given Name": v.get("givenName"),
                    "Surname": v.get("surname"),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df


@log
def list_group_owners(group: str | UUID) -> pd.DataFrame:
    """
    Shows a list of the owners of a group.

    This is a wrapper function for the following API: `List group owners <https://learn.microsoft.com/graph/api/group-list-owners>`_.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    group : str | uuid.UUID
        The group name or ID.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of the owners of a group.
    """

    group_id = resolve_group_id(group)

    result = _base_api(
        request=f"groups/{group_id}/owners", client="graph", uses_pagination=True
    )

    columns = {
        "Owner Id": "string",
        "Owner Name": "string",
        "User Principal Name": "string",
        "Mail": "string",
        "Job Title": "string",
        "Office Location": "string",
        "Mobile Phone": "string",
        "Business Phones": "string",
        "Preferred Language": "string",
        "Given Name": "string",
        "Surname": "string",
    }

    df = _create_dataframe(columns=columns)

    rows = []
    for r in result:
        for v in r.get("value", []):
            rows.append(
                {
                    "Owner Id": v.get("id"),
                    "Owner Name": v.get("displayName"),
                    "User Principal Name": v.get("userPrincipalName"),
                    "Mail": v.get("mail"),
                    "Job Title": v.get("jobTitle"),
                    "Office Location": v.get("officeLocation"),
                    "Mobile Phone": v.get("mobilePhone"),
                    "Business Phones": str(v.get("businessPhones")),
                    "Preferred Language": v.get("preferredLanguage"),
                    "Given Name": v.get("givenName"),
                    "Surname": v.get("surname"),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df


@log
def _base_add_to_group(
    group: str | UUID,
    object: str | UUID,
    object_type: Literal["members", "owners"],
):

    from sempy_labs.graph._users import resolve_user_id

    object_list = []

    if isinstance(object, str):
        object = [object]

    group_id = resolve_group_id(group)
    url = f"groups/{group_id}/{object_type}/$ref"

    for m in object:
        if _is_valid_uuid(m):
            member_id = m
        else:
            member_id = resolve_user_id(m)
        if object_type == "members":
            object_list.append(
                f"https://graph.microsoft.com/v1.0/directoryObjects/{member_id}"
            )
        else:
            object_list.append(f"https://graph.microsoft.com/v1.0/users/{member_id}")

    # Must submit one request for each owner. Members can be sent in a single request.
    if object_type == "members":
        payload = {"members@odata.bind": object_list}

        _base_api(
            request=url,
            client="graph",
            payload=payload,
            method="post",
            status_codes=204,
        )

    else:
        for o in object_list:
            payload = {"odata.id": o}
            _base_api(
                request=url,
                client="graph",
                payload=payload,
                method="post",
                status_codes=204,
            )

    print(
        f"{icons.green_dot} The {object} {object_type[:-1]}(s) have been added to the '{group}' group."
    )


@log
def add_group_members(
    group: str | UUID,
    user: str | UUID | List[str | UUID],
):
    """
    Adds a member to a group.

    This is a wrapper function for the following API: `Add members <https://learn.microsoft.com/graph/api/group-post-members>`_.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    group : str | uuid.UUID
        The group name or ID.
    user : str | uuid.UUID
        The user ID or user principal name.
    """

    _base_add_to_group(group=group, object=user, object_type="members")


@log
def add_group_owners(
    group: str | UUID,
    user: str | UUID | List[str | UUID],
):
    """
    Adds an owner to a group.

    This is a wrapper function for the following API: `Add owners <https://learn.microsoft.com/graph/api/group-post-owners>`_.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    group : str | uuid.UUID
        The group name or ID.
    user : str | uuid.UUID
        The user ID or user principal name.
    """

    _base_add_to_group(group=group, object=user, object_type="owners")


@log
def renew_group(group: str | UUID):
    """
    Renews the group.

    This is a wrapper function for the following API: `Renew group <https://learn.microsoft.com/graph/api/group-post-renew>`_.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    group : str | uuid.UUID
        The group name or ID.
    """

    group_id = resolve_group_id(group)

    _base_api(
        request=f"groups/{group_id}/renew",
        client="graph",
        method="post",
        status_codes=204,
    )

    print(f"{icons.green_dot} The '{group}' group has been renewed.")


@log
def create_group(
    display_name: str,
    description: Optional[str] = None,
    mail_enabled: bool = False,
    security_enabled: bool = True,
    mail_nickname: str = None,
    owners: Optional[str | UUID | List[str | UUID]] = None,
    members: Optional[str | UUID | List[str | UUID]] = None,
):
    """
    Creates a new group.

    This is a wrapper function for the following API: `Create group <https://learn.microsoft.com/graph/api/group-post-groups>`_.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    display_name : str
        The name of the group.
    description : str, optional
        The description of the group.
    mail_enabled : bool, default=False
        Whether the group is mail-enabled.
    security_enabled : bool, default=True
        Whether the group is security-enabled.
    mail_nickname : str, default=None
        The mail alias for the group.
    owners : str | uuid.UUID | List[str | uuid.UUID], default=None
        The owners of the group.
    members : str | uuid.UUID | List[str | uuid.UUID], default=None
        The members of the group.
    """
    from sempy_labs.graph._users import resolve_user_id

    payload = {
        "displayName": display_name,
        "description": description,
        "mailEnabled": mail_enabled,
        "securityEnabled": security_enabled,
        "mailNickname": mail_nickname,
    }

    if owners:
        if isinstance(owners, str):
            owners = [owners]
        user_list = []
        for o in owners:
            user_id = resolve_user_id(o)
            user_list.append(f"https://graph.microsoft.com/v1.0/users/{user_id}")
        payload["owners@odata.bind"] = user_list
    if members:
        if isinstance(members, str):
            members = [members]
        user_list = []
        for m in members:
            user_id = resolve_user_id(m)
            user_list.append(f"https://graph.microsoft.com/v1.0/users/{user_id}")
        payload["members@odata.bind"] = user_list

    _base_api(
        request="groups",
        client="graph",
        payload=payload,
        method="post",
        status_codes=201,
    )

    print(f"{icons.green_dot} The '{display_name}' group has been created.")


@log
def delete_group(group: str | UUID):
    """
    Deletes a group.

    This is a wrapper function for the following API: `Delete group <https://learn.microsoft.com/graph/api/group-delete>`_.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    group : str | uuid.UUID
        The group name or ID.
    """

    group_id = resolve_group_id(group)

    _base_api(
        request=f"groups/{group_id}",
        client="graph",
        status_codes=204,
        method="delete",
    )

    print(f"{icons.green_dot} The '{group}' group has been deleted successfully.")


@log
def update_group(
    group: str | UUID,
    display_name: Optional[str] = None,
    mail_nickname: Optional[str] = None,
    description: Optional[str] = None,
    security_enabled: Optional[bool] = None,
):
    """
    Updates a group's properties.

    This is a wrapper function for the following API: `Update group <https://learn.microsoft.com/en-us/graph/api/group-update>`_.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    group : str | uuid.UUID
        The group name or ID.
    display_name : str, default=None
        The new display name for the group.
    mail_nickname : str, default=None
        The new mail nickname for the group.
    description : str, default=None
        The new description for the group.
    security_enabled : bool, default=None
        Whether the group is security-enabled.
    """

    group_id = resolve_group_id(group)

    payload = {}
    if display_name:
        payload["displayName"] = display_name
    if mail_nickname:
        payload["mailNickname"] = mail_nickname
    if description:
        payload["description"] = description
    if security_enabled is not None and isinstance(security_enabled, bool):
        payload["securityEnabled"] = security_enabled

    if not payload:
        print(f"{icons.info} No properties to update.")
        return

    _base_api(
        request=f"groups/{group_id}",
        client="graph",
        status_codes=204,
        payload=payload,
        method="patch",
    )

    print(f"{icons.green_dot} The '{group}' group has been updated successfully.")



from uuid import UUID
from typing import Optional, List
import pandas as pd
from sempy_labs._helper_functions import (
    _create_dataframe,
    _base_api,
    _update_dataframe_datatypes,
    resolve_workspace_name_and_id,
    resolve_dataset_name_and_id,
    delete_item,
    resolve_dataset_id,
    resolve_workspace_id,
)
import sempy_labs._icons as icons
import re
from sempy._utils._log import log


@log
def get_semantic_model_refresh_schedule(
    dataset: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Gets the refresh schedule for the specified dataset from the specified workspace.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        Shows the refresh schedule for the specified dataset from the specified workspace.
    """

    workspace_id = resolve_workspace_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace)

    columns = {
        "Days": "str",
        "Times": "str",
        "Enabled": "bool",
        "Local Time Zone Id": "str",
        "Notify Option": "str",
    }

    column_map = {
        "days": "Days",
        "times": "Times",
        "enabled": "Enabled",
        "localTimeZoneId": "Local Time Zone Id",
        "notifyOption": "Notify Option",
    }

    df = _create_dataframe(columns)

    result = _base_api(
        request=f"/v1.0/myorg/groups/{workspace_id}/datasets/{dataset_id}/refreshSchedule",
        client="fabric_sp",
    ).json()

    df = (
        pd.json_normalize(result)
        .drop(columns=["@odata.context"], errors="ignore")
        .rename(columns=column_map)
    )

    _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


@log
def enable_semantic_model_scheduled_refresh(
    dataset: str | UUID,
    workspace: Optional[str | UUID] = None,
    enable: bool = True,
):
    """
    Enables the scheduled refresh for the specified dataset from the specified workspace.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    enable : bool, default=True
        If True, enables the scheduled refresh.
        If False, disables the scheduled refresh.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace)

    df = get_semantic_model_refresh_schedule(dataset=dataset, workspace=workspace)
    status = df["Enabled"].iloc[0]

    if enable and status:
        print(
            f"{icons.info} Scheduled refresh for the '{dataset_name}' within the '{workspace_name}' workspace is already enabled."
        )
    elif not enable and not status:
        print(
            f"{icons.info} Scheduled refresh for the '{dataset_name}' within the '{workspace_name}' workspace is already disabled."
        )
    else:
        payload = {"value": {"enabled": enable}}

        _base_api(
            request=f"/v1.0/myorg/groups/{workspace_id}/datasets/{dataset_id}/refreshSchedule",
            method="patch",
            payload=payload,
            client="fabric_sp",
        )

        print(
            f"{icons.green_dot} Scheduled refresh for the '{dataset_name}' within the '{workspace_name}' workspace has been enabled."
        )


@log
def delete_semantic_model(dataset: str | UUID, workspace: Optional[str | UUID] = None):
    """
    Deletes a semantic model.

    This is a wrapper function for the following API: `Items - Delete Semantic Model <https://learn.microsoft.com/rest/api/fabric/semanticmodel/items/delete-semantic-model>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    dataset: str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    delete_item(item=dataset, type="SemanticModel", workspace=workspace)


@log
def update_semantic_model_refresh_schedule(
    dataset: str | UUID,
    days: Optional[str | List[str]] = None,
    times: Optional[str | List[str]] = None,
    time_zone: Optional[str] = None,
    workspace: Optional[str | UUID] = None,
):
    """
    Updates the refresh schedule for the specified dataset from the specified workspace.

    This is a wrapper function for the following API: `Datasets - Update Refresh Schedule In Group <https://learn.microsoft.com/rest/api/power-bi/datasets/update-refresh-schedule-in-group>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    days : str | list[str], default=None
        The days of the week to refresh the dataset.
        Valid values are: "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday".
        Defaults to None which means the refresh schedule will not be updated.
    times : str | list[str], default=None
        The times of the day to refresh the dataset.
        Valid format is "HH:MM" (24-hour format).
        Defaults to None which means the refresh schedule will not be updated.
    time_zone : str, default=None
        The time zone to use for the refresh schedule.
        Defaults to None which means the refresh schedule will not be updated.
    workspace : str | uuid.UUID, default=None
        The workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace)

    payload = {"value": {}}

    def is_valid_time_format(time_str):
        pattern = r"^(?:[01]\d|2[0-3]):[0-5]\d$"
        return re.match(pattern, time_str) is not None

    weekdays = [
        "Monday",
        "Tuesday",
        "Wednesday",
        "Thursday",
        "Friday",
        "Sunday",
        "Saturday",
    ]
    if days:
        if isinstance(days, str):
            days = [days]
            for i in range(len(days)):
                days[i] = days[i].capitalize()
                if days[i] not in weekdays:
                    raise ValueError(
                        f"{icons.red_dot} Invalid day '{days[i]}'. Valid days are: {weekdays}"
                    )
        payload["value"]["days"] = days
    if times:
        if isinstance(times, str):
            times = [times]
            for i in range(len(times)):
                if not is_valid_time_format(times[i]):
                    raise ValueError(
                        f"{icons.red_dot} Invalid time '{times[i]}'. Valid time format is 'HH:MM' (24-hour format)."
                    )
        payload["value"]["times"] = times
    if time_zone:
        payload["value"]["localTimeZoneId"] = time_zone

    if not payload.get("value"):
        print(
            f"{icons.info} No changes were made to the refresh schedule for the '{dataset_name}' within the '{workspace_name}' workspace."
        )
        return

    _base_api(
        request=f"/v1.0/myorg/groups/{workspace_id}/datasets/{dataset_id}/refreshSchedule",
        method="patch",
        client="fabric_sp",
        payload=payload,
    )

    print(
        f"{icons.green_dot} Refresh schedule for the '{dataset_name}' within the '{workspace_name}' workspace has been updated."
    )


@log
def list_semantic_model_datasources(
    dataset: str | UUID,
    workspace: Optional[str | UUID] = None,
    expand_details: bool = True,
) -> pd.DataFrame:
    """
    Lists the data sources for the specified semantic model.

    This is a wrapper function for the following API: `Datasets - Get Datasources In Group <https://learn.microsoft.com/rest/api/power-bi/datasets/get-datasources-in-group>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    expand_details : bool, default=True
        If True, expands the connection details for each data source.

    Returns
    -------
    pandas.DataFrame
        DataFrame containing the data sources for the specified semantic model.
    """

    workspace_id = resolve_workspace_id(workspace)
    dataset_id = resolve_dataset_id(dataset, workspace_id)

    if expand_details:
        columns = {
            "Datasource Type": "str",
            "Connection Server": "str",
            "Connection Database": "str",
            "Connection Path": "str",
            "Connection Account": "str",
            "Connection Domain": "str",
            "Connection Kind": "str",
            "Connection Email Address": "str",
            "Connection URL": "str",
            "Connection Class Info": "str",
            "Connection Login Server": "str",
            "Datasource Id": "str",
            "Gateway Id": "str",
        }
    else:
        columns = {
            "Datasource Type": "str",
            "Connection Details": "str",
            "Datasource Id": "str",
            "Gateway Id": "str",
        }

    df = _create_dataframe(columns)

    response = _base_api(
        request=f"/v1.0/myorg/groups/{workspace_id}/datasets/{dataset_id}/datasources",
        client="fabric_sp",
    )

    rows = []
    for item in response.json().get("value", []):
        ds_type = item.get("datasourceType")
        conn_details = item.get("connectionDetails", {})
        ds_id = item.get("datasourceId")
        gateway_id = item.get("gatewayId")
        if expand_details:
            rows.append(
                {
                    "Datasource Type": ds_type,
                    "Connection Server": conn_details.get("server"),
                    "Connection Database": conn_details.get("database"),
                    "Connection Path": conn_details.get("path"),
                    "Connection Account": conn_details.get("account"),
                    "Connection Domain": conn_details.get("domain"),
                    "Connection Kind": conn_details.get("kind"),
                    "Connection Email Address": conn_details.get("emailAddress"),
                    "Connection URL": conn_details.get("url"),
                    "Connection Class Info": conn_details.get("classInfo"),
                    "Connection Login Server": conn_details.get("loginServer"),
                    "Datasource Id": ds_id,
                    "Gateway Id": gateway_id,
                }
            )
        else:
            rows.append(
                {
                    "Datasource Type": ds_type,
                    "Connection Details": conn_details,
                    "Datasource Id": ds_id,
                    "Gateway Id": gateway_id,
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df


@log
def bind_semantic_model_connection(
    dataset: str | UUID,
    connection_id: UUID,
    connectivity_type: str,
    connection_type: str,
    connection_path: str,
    workspace: Optional[str | UUID] = None,
):
    """
    Binds a semantic model data source reference to a data connection.
    This API can also be used to unbind data source references.

    This is a wrapper function for the following API: `Items - Bind Semantic Model Connection <https://learn.microsoft.com/rest/api/fabric/semanticmodel/items/bind-semantic-model-connection>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    connection_id : uuid.UUID
        The object ID of the connection.
    connectivity_type : str
        The connectivity type of the connection. Additional connectivity types may be added over time.
    connection_type : str
        The `type <https://learn.microsoft.com/rest/api/fabric/semanticmodel/items/bind-semantic-model-connection?tabs=HTTP#connectivitytype>`_ of the connection.
    connection_path : str
        The path of the connection.
    workspace : str | uuid.UUID, default=None
        The workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(
        dataset=dataset, workspace=workspace_id
    )

    payload = {
        "connectionBinding": {
            "id": str(connection_id),
            "connectivityType": connectivity_type,
            "connectionDetails": {
                "type": connection_type,
                "path": connection_path,
            },
        }
    }

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/semanticModels/{dataset_id}/bindConnection",
        method="post",
        client="fabric_sp",
        payload=payload,
    )

    print(
        f"{icons.green_dot} Connection '{connection_id}' has been bound to the '{dataset_name}' semantic model within the '{workspace_name}' workspace."
    )


@log
def unbind_semantic_model_connection(
    dataset: str | UUID,
    connection_type: str,
    connection_path: str,
    workspace: Optional[str | UUID] = None,
):
    """
    Unbinds a semantic model data source reference to a data connection.

    This is a wrapper function for the following API: `Items - Bind Semantic Model Connection <https://learn.microsoft.com/rest/api/fabric/semanticmodel/items/bind-semantic-model-connection>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    connection_type : str
        The `type <https://learn.microsoft.com/rest/api/fabric/semanticmodel/items/bind-semantic-model-connection?tabs=HTTP#connectivitytype>`_ of the connection.
    connection_path : str
        The path of the connection.
    workspace : str | uuid.UUID, default=None
        The workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(
        dataset=dataset, workspace=workspace_id
    )

    payload = {
        "connectionBinding": {
            "connectivityType": "None",
            "connectionDetails": {
                "type": connection_type,
                "path": connection_path,
            },
        }
    }

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/semanticModels/{dataset_id}/bindConnection",
        method="post",
        client="fabric_sp",
        payload=payload,
    )

    print(
        f"{icons.green_dot} The '{dataset_name}' semantic model within the '{workspace_name}' workspace has been unbound from its connection."
    )
